@@ -33,10 +33,9 @@
 #endif
 #include <ctype.h>
 
-#include "ruby/ruby.h"
+#include "internal.h"
 #include "ruby/io.h"
 #include "ruby/util.h"
-#include "internal.h"
 
 #include <signal.h>
 #ifdef HAVE_SYS_STROPTS_H
@@ -183,7 +182,7 @@ establishShell(int argc, VALUE *argv, struct pty_info *info,
 
     carg.execarg_obj = rb_execarg_new(argc, argv, 1);
     carg.eargp = rb_execarg_get(carg.execarg_obj);
-    rb_execarg_fixup(carg.execarg_obj);
+    rb_execarg_parent_start(carg.execarg_obj);
 
     getDevice(&master, &slave, SlaveName, 0);
 
@@ -197,12 +196,14 @@ establishShell(int argc, VALUE *argv, struct pty_info *info,
 	int e = errno;
 	close(master);
 	close(slave);
+        rb_execarg_parent_end(carg.execarg_obj);
 	errno = e;
 	if (status) rb_jump_tag(status);
 	rb_sys_fail(errbuf[0] ? errbuf : "fork failed");
     }
 
     close(slave);
+    rb_execarg_parent_end(carg.execarg_obj);
 
     info->child_pid = pid;
     info->fd = master;
@@ -504,6 +505,14 @@ pty_close_pty(VALUE assoc)
  * +slave_file+::   the slave of the pty, as a File.  The path to the
  *		    terminal device is available via +slave_file.path+
  *
+ * IO#raw! is usable to disable newline conversions:
+ *
+ *   require 'io/console'
+ *   PTY.open {|m, s|
+ *     s.raw!
+ *     ...
+ *   }
+ *
  */
 static VALUE
 pty_open(VALUE klass)
@@ -538,6 +547,11 @@ pty_open(VALUE klass)
 static VALUE
 pty_detach_process(struct pty_info *info)
 {
+#ifdef WNOHANG
+    int st;
+    if (rb_waitpid(info->child_pid, &st, WNOHANG) <= 0)
+	return Qnil;
+#endif
     rb_detach_process(info->child_pid);
     return Qnil;
 }
@@ -565,7 +579,7 @@ pty_detach_process(struct pty_info *info)
  *
  * In the block form these same values will be yielded to the block:
  *
- * +r+:: A readable IO that that contains the command's
+ * +r+:: A readable IO that contains the command's
  *       standard output and standard error
  * +w+:: A writable IO that is the command's standard input
  * +pid+:: The process identifier for the command.
@@ -585,11 +599,11 @@ pty_getpty(int argc, VALUE *argv, VALUE self)
 
     establishShell(argc, argv, &info, SlaveName);
 
-    rfptr->mode = rb_io_mode_flags("r");
+    rfptr->mode = rb_io_modestr_fmode("r");
     rfptr->fd = info.fd;
     rfptr->pathv = rb_obj_freeze(rb_str_new_cstr(SlaveName));
 
-    wfptr->mode = rb_io_mode_flags("w") | FMODE_SYNC;
+    wfptr->mode = rb_io_modestr_fmode("w") | FMODE_SYNC;
     wfptr->fd = rb_cloexec_dup(info.fd);
     if (wfptr->fd == -1)
         rb_sys_fail("dup()");
@@ -719,7 +733,7 @@ static VALUE cPTY;
  *   # The result of read operation when pty slave is closed is platform
  *   # dependent.
  *   ret = begin
- *           m.gets          # FreeBSD returns nil.
+ *           master.gets     # FreeBSD returns nil.
  *         rescue Errno::EIO # GNU/Linux raises EIO.
  *           nil
  *         end
@@ -740,10 +754,10 @@ static VALUE cPTY;
  */
 
 void
-Init_pty()
+Init_pty(void)
 {
     cPTY = rb_define_module("PTY");
-    /* :nodoc */
+    /* :nodoc: */
     rb_define_module_function(cPTY,"getpty",pty_getpty,-1);
     rb_define_module_function(cPTY,"spawn",pty_getpty,-1);
     rb_define_singleton_method(cPTY,"check",pty_check,-1);
