@@ -1,4 +1,5 @@
 # coding: US-ASCII
+# frozen_string_literal: false
 require 'test/unit'
 
 class TestRegexp < Test::Unit::TestCase
@@ -550,6 +551,16 @@ class TestRegexp < Test::Unit::TestCase
     assert_nothing_raised { $= = nil }
   end
 
+  def test_KCODE_warning
+    assert_warning(/variable \$KCODE is no longer effective; ignored/) { $KCODE = nil }
+    assert_warning(/variable \$KCODE is no longer effective/) { $KCODE = nil }
+  end
+
+  def test_ignorecase_warning
+    assert_warning(/variable \$= is no longer effective; ignored/) { $= = nil }
+    assert_warning(/variable \$= is no longer effective/) { $= }
+  end
+
   def test_match_setter
     /foo/ =~ "foo"
     m = $~
@@ -567,11 +578,9 @@ class TestRegexp < Test::Unit::TestCase
     key = "\u{3042}"
     [Encoding::UTF_8, Encoding::Shift_JIS, Encoding::EUC_JP].each do |enc|
       idx = key.encode(enc)
-      test.call {|m|
-        EnvUtil.with_default_external(enc) do
-          assert_raise_with_message(IndexError, /#{idx}/, bug10877) {m[idx]}
-        end
-      }
+      EnvUtil.with_default_external(enc) do
+        test.call {|m| assert_raise_with_message(IndexError, /#{idx}/, bug10877) {m[idx]} }
+      end
     end
     test.call {|m| assert_equal(/a/, m.regexp) }
     test.call {|m| assert_equal("abc", m.string) }
@@ -615,19 +624,6 @@ class TestRegexp < Test::Unit::TestCase
     assert_equal(3, "foobarbaz\u3042".rindex(/b../n, 5))
   end
 
-  def test_taint
-    m = Thread.new do
-      "foo"[/foo/]
-      $SAFE = 3
-      /foo/.match("foo")
-    end.value
-    assert_predicate(m, :tainted?)
-    assert_nothing_raised('[ruby-core:26137]') {
-      m = proc {$SAFE = 3; %r"#{ }"o}.call
-    }
-    assert_predicate(m, :tainted?)
-  end
-
   def assert_regexp(re, ss, fs = [], msg = nil)
     re = Regexp.new(re) unless re.is_a?(Regexp)
     ss = [ss] unless ss.is_a?(Array)
