@@ -1,5 +1,5 @@
+# frozen_string_literal: false
 require 'test/unit'
-require 'rational'
 require 'delegate'
 require 'timeout'
 require 'delegate'
@@ -14,6 +14,29 @@ class TestTime < Test::Unit::TestCase
     $VERBOSE = @verbose
   end
 
+  def in_timezone(zone)
+    orig_zone = ENV['TZ']
+
+    ENV['TZ'] = zone
+    yield
+  ensure
+    ENV['TZ'] = orig_zone
+  end
+
+  def no_leap_seconds?
+    # 1972-06-30T23:59:60Z is the first leap second.
+    Time.utc(1972, 7, 1, 0, 0, 0) - Time.utc(1972, 6, 30, 23, 59, 59) == 1
+  end
+
+  def get_t2000
+    if no_leap_seconds?
+      # Sat Jan 01 00:00:00 UTC 2000
+      Time.at(946684800).gmtime
+    else
+      Time.utc(2000, 1, 1)
+    end
+  end
+
   def test_new
     assert_equal(Time.utc(2000,2,10), Time.new(2000,2,10, 11,0,0, 3600*11))
     assert_equal(Time.utc(2000,2,10), Time.new(2000,2,9, 13,0,0, -3600*11))
@@ -23,6 +46,7 @@ class TestTime < Test::Unit::TestCase
     tm = [2001,2,28,23,59,30]
     t = Time.new(*tm, "-12:00")
     assert_equal([2001,2,28,23,59,30,-43200], [t.year, t.month, t.mday, t.hour, t.min, t.sec, t.gmt_offset], bug4090)
+    assert_raise(ArgumentError) { Time.new(2000,1,1, 0,0,0, "+01:60") }
   end
 
   def test_time_add()
@@ -32,14 +56,14 @@ class TestTime < Test::Unit::TestCase
                  Time.utc(2000, 3, 21, 0, 30))
     assert_equal(0, (Time.at(1.1) + 0.9).usec)
 
-    assert((Time.utc(2000, 4, 1) + 24).utc?)
-    assert(!(Time.local(2000, 4, 1) + 24).utc?)
+    assert_predicate((Time.utc(2000, 4, 1) + 24), :utc?)
+    assert_not_predicate((Time.local(2000, 4, 1) + 24), :utc?)
 
     t = Time.new(2000, 4, 1, 0, 0, 0, "+01:00") + 24
-    assert(!t.utc?)
+    assert_not_predicate(t, :utc?)
     assert_equal(3600, t.utc_offset)
     t = Time.new(2000, 4, 1, 0, 0, 0, "+02:00") + 24
-    assert(!t.utc?)
+    assert_not_predicate(t, :utc?)
     assert_equal(7200, t.utc_offset)
   end
 
@@ -80,15 +104,13 @@ class TestTime < Test::Unit::TestCase
     assert_equal(31536000, Time.utc(1971, 1, 1, 0, 0, 0).tv_sec)
     assert_equal(78796799, Time.utc(1972, 6, 30, 23, 59, 59).tv_sec)
 
-    # 1972-06-30T23:59:60Z is the first leap second.
-    if Time.utc(1972, 7, 1, 0, 0, 0) - Time.utc(1972, 6, 30, 23, 59, 59) == 1
-      # no leap second.
+    if no_leap_seconds?
       assert_equal(78796800, Time.utc(1972, 7, 1, 0, 0, 0).tv_sec)
       assert_equal(78796801, Time.utc(1972, 7, 1, 0, 0, 1).tv_sec)
       assert_equal(946684800, Time.utc(2000, 1, 1, 0, 0, 0).tv_sec)
       assert_equal(0x7fffffff, Time.utc(2038, 1, 19, 3, 14, 7).tv_sec)
+      assert_equal(0x80000000, Time.utc(2038, 1, 19, 3, 14, 8).tv_sec)
     else
-      # leap seconds supported.
       assert_equal(2, Time.utc(1972, 7, 1, 0, 0, 0) - Time.utc(1972, 6, 30, 23, 59, 59))
       assert_equal(78796800, Time.utc(1972, 6, 30, 23, 59, 60).tv_sec)
       assert_equal(78796801, Time.utc(1972, 7, 1, 0, 0, 0).tv_sec)
@@ -150,6 +172,7 @@ class TestTime < Test::Unit::TestCase
     assert_equal(100000, Time.at(0.0001).nsec)
     assert_equal(10000, Time.at(0.00001).nsec)
     assert_equal(3000, Time.at(0.000003).nsec)
+    assert_equal(200, Time.at(0.0000002r).nsec)
     assert_equal(199, Time.at(0.0000002).nsec)
     assert_equal(10, Time.at(0.00000001).nsec)
     assert_equal(1, Time.at(0.000000001).nsec)
@@ -176,7 +199,7 @@ class TestTime < Test::Unit::TestCase
 
     t = Time.at(2**40 + "1/3".to_r, 9999999999999).utc
     assert_equal(36812, t.year)
-    
+
     t = Time.at(-0x3fff_ffff_ffff_ffff)
     assert_equal(-146138510344, t.year)
     t = Time.at(-0x4000_0000_0000_0000)
@@ -277,21 +300,71 @@ class TestTime < Test::Unit::TestCase
     assert_equal(29700, t2.utc_offset, bug)
   end
 
-  # Sat Jan 01 00:00:00 UTC 2000
-  T2000 = Time.at(946684800).gmtime
+  def test_marshal_zone
+    t = Time.utc(2013, 2, 24)
+    assert_equal('UTC', t.zone)
+    assert_equal('UTC', Marshal.load(Marshal.dump(t)).zone)
 
-  def test_security_error
-    assert_raise(SecurityError) do
-      Thread.new do
-        t = Time.gm(2000)
-        $SAFE = 4
-        t.localtime
-      end.join
+    in_timezone('JST-9') do
+      t = Time.local(2013, 2, 24)
+      assert_equal('JST', Time.local(2013, 2, 24).zone)
+      assert_equal('JST', Marshal.load(Marshal.dump(t)).zone)
+    end
+  end
+
+  def test_marshal_zone_gc
+    assert_separately(%w(--disable-gems), <<-'end;', timeout: 30)
+      ENV["TZ"] = "JST-9"
+      s = Marshal.dump(Time.now)
+      t = Marshal.load(s)
+      n = 0
+      done = 100000
+      while t.zone.dup == "JST" && n < done
+        n += 1
+      end
+      assert_equal done, n, "Bug #9652"
+      assert_equal "JST", t.zone, "Bug #9652"
+    end;
+  end
+
+  def test_marshal_to_s
+    t1 = Time.new(2011,11,8, 0,42,25, 9*3600)
+    t2 = Time.at(Marshal.load(Marshal.dump(t1)))
+    assert_equal("2011-11-08 00:42:25 +0900", t2.to_s,
+      "[ruby-dev:44827] [Bug #5586]")
+  end
+
+  Bug8795 = '[ruby-core:56648] [Bug #8795]'
+
+  def test_marshal_broken_offset
+    data = "\x04\bIu:\tTime\r\xEFF\x1C\x80\x00\x00\x00\x00\x06:\voffset"
+    t1 = t2 = nil
+    in_timezone('UTC') do
+      assert_nothing_raised(TypeError, ArgumentError, Bug8795) do
+        t1 = Marshal.load(data + "T")
+        t2 = Marshal.load(data + "\"\x0ebadoffset")
+      end
+      assert_equal(0, t1.utc_offset)
+      assert_equal(0, t2.utc_offset)
+    end
+  end
+
+  def test_marshal_broken_zone
+    data = "\x04\bIu:\tTime\r\xEFF\x1C\x80\x00\x00\x00\x00\x06:\tzone"
+    t1 = t2 = nil
+    in_timezone('UTC') do
+      assert_nothing_raised(TypeError, ArgumentError, Bug8795) do
+        t1 = Marshal.load(data + "T")
+        t2 = Marshal.load(data + "\"\b\0\0\0")
+      end
+      assert_equal('UTC', t1.zone)
+      assert_equal('UTC', t2.zone)
     end
   end
 
   def test_at3
-    assert_equal(T2000, Time.at(T2000))
+    t2000 = get_t2000
+    assert_equal(t2000, Time.at(t2000))
 #    assert_raise(RangeError) do
 #      Time.at(2**31-1, 1_000_000)
 #      Time.at(2**63-1, 1_000_000)
@@ -303,13 +376,14 @@ class TestTime < Test::Unit::TestCase
   end
 
   def test_utc_or_local
-    assert_equal(T2000, Time.gm(2000))
-    assert_equal(T2000, Time.gm(0, 0, 0, 1, 1, 2000, :foo, :bar, false, :baz))
-    assert_equal(T2000, Time.gm(2000, "jan"))
-    assert_equal(T2000, Time.gm(2000, "1"))
-    assert_equal(T2000, Time.gm(2000, 1, 1, 0, 0, 0, 0))
-    assert_equal(T2000, Time.gm(2000, 1, 1, 0, 0, 0, "0"))
-    assert_equal(T2000, Time.gm(2000, 1, 1, 0, 0, "0", :foo, :foo))
+    t2000 = get_t2000
+    assert_equal(t2000, Time.gm(2000))
+    assert_equal(t2000, Time.gm(0, 0, 0, 1, 1, 2000, :foo, :bar, false, :baz))
+    assert_equal(t2000, Time.gm(2000, "jan"))
+    assert_equal(t2000, Time.gm(2000, "1"))
+    assert_equal(t2000, Time.gm(2000, 1, 1, 0, 0, 0, 0))
+    assert_equal(t2000, Time.gm(2000, 1, 1, 0, 0, 0, "0"))
+    assert_equal(t2000, Time.gm(2000, 1, 1, 0, 0, "0", :foo, :foo))
     assert_raise(ArgumentError) { Time.gm(2000, 1, 1, 0, 0, -1, :foo, :foo) }
     assert_raise(ArgumentError) { Time.gm(2000, 1, 1, 0, 0, -1.0, :foo, :foo) }
     assert_raise(RangeError) do
@@ -317,6 +391,7 @@ class TestTime < Test::Unit::TestCase
     end
     assert_raise(ArgumentError) { Time.gm(2000, 1, 1, 0, 0, -(2**31), :foo, :foo) }
     o = Object.new
+    def o.to_int; 0; end
     def o.to_r; nil; end
     assert_raise(TypeError) { Time.gm(2000, 1, 1, 0, 0, o, :foo, :foo) }
     def o.to_r; ""; end
@@ -329,7 +404,7 @@ class TestTime < Test::Unit::TestCase
     assert_raise(ArgumentError) { Time.gm(2000, 13) }
 
     t = Time.local(2000)
-    assert_equal(t.gmt_offset, T2000 - t)
+    assert_equal(t.gmt_offset, t2000 - t)
 
     assert_equal(-4427700000, Time.utc(-4427700000,12,1).year)
     assert_equal(-2**30+10, Time.utc(-2**30+10,1,1).year)
@@ -343,52 +418,77 @@ class TestTime < Test::Unit::TestCase
       }
     }
     assert_raise(ArgumentError) { m.sleep(-1) }
+    assert_raise(TypeError) { m.sleep("") }
+    assert_raise(TypeError) { sleep("") }
+    obj = eval("class C\u{1f5ff}; self; end").new
+    assert_raise_with_message(TypeError, /C\u{1f5ff}/) {m.sleep(obj)}
+    assert_raise_with_message(TypeError, /C\u{1f5ff}/) {sleep(obj)}
   end
 
   def test_to_f
-    assert_equal(946684800.0, T2000.to_f)
+    t2000 = Time.at(946684800).gmtime
+    assert_equal(946684800.0, t2000.to_f)
+  end
+
+  def test_to_f_accuracy
+    # https://bugs.ruby-lang.org/issues/10135#note-1
+    f = 1381089302.195
+    assert_equal(f, Time.at(f).to_f, "[ruby-core:64373] [Bug #10135] note-1")
   end
 
   def test_cmp
-    assert_equal(-1, T2000 <=> Time.gm(2001))
-    assert_equal(1, T2000 <=> Time.gm(1999))
-    assert_nil(T2000 <=> 0)
+    t2000 = get_t2000
+    assert_equal(-1, t2000 <=> Time.gm(2001))
+    assert_equal(1, t2000 <=> Time.gm(1999))
+    assert_nil(t2000 <=> 0)
   end
 
   def test_eql
-    assert(T2000.eql?(T2000))
-    assert(!T2000.eql?(Time.gm(2001)))
+    t2000 = get_t2000
+    assert_operator(t2000, :eql?, t2000)
+    assert_not_operator(t2000, :eql?, Time.gm(2001))
   end
 
   def test_utc_p
-    assert(Time.gm(2000).gmt?)
-    assert(!Time.local(2000).gmt?)
-    assert(!Time.at(0).gmt?)
+    assert_predicate(Time.gm(2000), :gmt?)
+    assert_not_predicate(Time.local(2000), :gmt?)
+    assert_not_predicate(Time.at(0), :gmt?)
   end
 
   def test_hash
-    assert_kind_of(Integer, T2000.hash)
+    t2000 = get_t2000
+    assert_kind_of(Integer, t2000.hash)
+  end
+
+  def test_reinitialize
+    bug8099 = '[ruby-core:53436] [Bug #8099]'
+    t2000 = get_t2000
+    assert_raise(TypeError, bug8099) {
+      t2000.send(:initialize, 2013, 03, 14)
+    }
+    assert_equal(get_t2000, t2000, bug8099)
   end
 
   def test_init_copy
-    assert_equal(T2000, T2000.dup)
+    t2000 = get_t2000
+    assert_equal(t2000, t2000.dup)
     assert_raise(TypeError) do
-      T2000.instance_eval { initialize_copy(nil) }
+      t2000.instance_eval { initialize_copy(nil) }
     end
   end
 
   def test_localtime_gmtime
     assert_nothing_raised do
       t = Time.gm(2000)
-      assert(t.gmt?)
+      assert_predicate(t, :gmt?)
       t.localtime
-      assert(!t.gmt?)
+      assert_not_predicate(t, :gmt?)
       t.localtime
-      assert(!t.gmt?)
+      assert_not_predicate(t, :gmt?)
       t.gmtime
-      assert(t.gmt?)
+      assert_predicate(t, :gmt?)
       t.gmtime
-      assert(t.gmt?)
+      assert_predicate(t, :gmt?)
     end
 
     t1 = Time.gm(2000)
@@ -397,6 +497,7 @@ class TestTime < Test::Unit::TestCase
     t3 = t1.getlocal("-02:00")
     assert_equal(t1, t3)
     assert_equal(-7200, t3.utc_offset)
+    assert_equal([1999, 12, 31, 22, 0, 0], [t3.year, t3.mon, t3.mday, t3.hour, t3.min, t3.sec])
     t1.localtime
     assert_equal(t1, t2)
     assert_equal(t1.gmt?, t2.gmt?)
@@ -415,22 +516,42 @@ class TestTime < Test::Unit::TestCase
   end
 
   def test_asctime
-    assert_equal("Sat Jan  1 00:00:00 2000", T2000.asctime)
+    t2000 = get_t2000
+    assert_equal("Sat Jan  1 00:00:00 2000", t2000.asctime)
+    assert_equal(Encoding::US_ASCII, t2000.asctime.encoding)
     assert_kind_of(String, Time.at(0).asctime)
   end
 
   def test_to_s
-    assert_equal("2000-01-01 00:00:00 UTC", T2000.to_s)
+    t2000 = get_t2000
+    assert_equal("2000-01-01 00:00:00 UTC", t2000.to_s)
+    assert_equal(Encoding::US_ASCII, t2000.to_s.encoding)
     assert_kind_of(String, Time.at(946684800).getlocal.to_s)
     assert_equal(Time.at(946684800).getlocal.to_s, Time.at(946684800).to_s)
   end
 
+  def assert_zone_encoding(time)
+    zone = time.zone
+    assert_predicate(zone, :valid_encoding?)
+    if zone.ascii_only?
+      assert_equal(Encoding::US_ASCII, zone.encoding)
+    else
+      enc = Encoding.default_internal || Encoding.find('locale')
+      assert_equal(enc, zone.encoding)
+    end
+  end
+
+  def test_zone
+    assert_zone_encoding Time.now
+  end
+
   def test_plus_minus_succ
-    # assert_raise(RangeError) { T2000 + 10000000000 }
-    # assert_raise(RangeError)  T2000 - 3094168449 }
-    # assert_raise(RangeError) { T2000 + 1200798848 }
-    assert_raise(TypeError) { T2000 + Time.now }
-    assert_equal(T2000 + 1, T2000.succ)
+    t2000 = get_t2000
+    # assert_raise(RangeError) { t2000 + 10000000000 }
+    # assert_raise(RangeError)  t2000 - 3094168449 }
+    # assert_raise(RangeError) { t2000 + 1200798848 }
+    assert_raise(TypeError) { t2000 + Time.now }
+    assert_equal(t2000 + 1, t2000.succ)
   end
 
   def test_plus_type
@@ -459,26 +580,27 @@ class TestTime < Test::Unit::TestCase
   end
 
   def test_readers
-    assert_equal(0, T2000.sec)
-    assert_equal(0, T2000.min)
-    assert_equal(0, T2000.hour)
-    assert_equal(1, T2000.mday)
-    assert_equal(1, T2000.mon)
-    assert_equal(2000, T2000.year)
-    assert_equal(6, T2000.wday)
-    assert_equal(1, T2000.yday)
-    assert_equal(false, T2000.isdst)
-    assert_equal("UTC", T2000.zone)
-    assert_equal(Encoding.find("locale"), T2000.zone.encoding)
-    assert_equal(0, T2000.gmt_offset)
-    assert(!T2000.sunday?)
-    assert(!T2000.monday?)
-    assert(!T2000.tuesday?)
-    assert(!T2000.wednesday?)
-    assert(!T2000.thursday?)
-    assert(!T2000.friday?)
-    assert(T2000.saturday?)
-    assert_equal([0, 0, 0, 1, 1, 2000, 6, 1, false, "UTC"], T2000.to_a)
+    t2000 = get_t2000
+    assert_equal(0, t2000.sec)
+    assert_equal(0, t2000.min)
+    assert_equal(0, t2000.hour)
+    assert_equal(1, t2000.mday)
+    assert_equal(1, t2000.mon)
+    assert_equal(2000, t2000.year)
+    assert_equal(6, t2000.wday)
+    assert_equal(1, t2000.yday)
+    assert_equal(false, t2000.isdst)
+    assert_equal("UTC", t2000.zone)
+    assert_zone_encoding(t2000)
+    assert_equal(0, t2000.gmt_offset)
+    assert_not_predicate(t2000, :sunday?)
+    assert_not_predicate(t2000, :monday?)
+    assert_not_predicate(t2000, :tuesday?)
+    assert_not_predicate(t2000, :wednesday?)
+    assert_not_predicate(t2000, :thursday?)
+    assert_not_predicate(t2000, :friday?)
+    assert_predicate(t2000, :saturday?)
+    assert_equal([0, 0, 0, 1, 1, 2000, 6, 1, false, "UTC"], t2000.to_a)
 
     t = Time.at(946684800).getlocal
     assert_equal(t.sec, Time.at(946684800).sec)
@@ -491,7 +613,7 @@ class TestTime < Test::Unit::TestCase
     assert_equal(t.yday, Time.at(946684800).yday)
     assert_equal(t.isdst, Time.at(946684800).isdst)
     assert_equal(t.zone, Time.at(946684800).zone)
-    assert_equal(Encoding.find("locale"), Time.at(946684800).zone.encoding)
+    assert_zone_encoding(Time.at(946684800))
     assert_equal(t.gmt_offset, Time.at(946684800).gmt_offset)
     assert_equal(t.sunday?, Time.at(946684800).sunday?)
     assert_equal(t.monday?, Time.at(946684800).monday?)
@@ -504,51 +626,68 @@ class TestTime < Test::Unit::TestCase
   end
 
   def test_strftime
+    t2000 = get_t2000
     t = Time.at(946684800).getlocal
-    assert_equal("Sat", T2000.strftime("%a"))
-    assert_equal("Saturday", T2000.strftime("%A"))
-    assert_equal("Jan", T2000.strftime("%b"))
-    assert_equal("January", T2000.strftime("%B"))
-    assert_kind_of(String, T2000.strftime("%c"))
-    assert_equal("01", T2000.strftime("%d"))
-    assert_equal("00", T2000.strftime("%H"))
-    assert_equal("12", T2000.strftime("%I"))
-    assert_equal("001", T2000.strftime("%j"))
-    assert_equal("01", T2000.strftime("%m"))
-    assert_equal("00", T2000.strftime("%M"))
-    assert_equal("AM", T2000.strftime("%p"))
-    assert_equal("00", T2000.strftime("%S"))
-    assert_equal("00", T2000.strftime("%U"))
-    assert_equal("00", T2000.strftime("%W"))
-    assert_equal("6", T2000.strftime("%w"))
-    assert_equal("01/01/00", T2000.strftime("%x"))
-    assert_equal("00:00:00", T2000.strftime("%X"))
-    assert_equal("00", T2000.strftime("%y"))
-    assert_equal("2000", T2000.strftime("%Y"))
-    assert_equal("UTC", T2000.strftime("%Z"))
-    assert_equal("%", T2000.strftime("%%"))
-    assert_equal("0", T2000.strftime("%-S"))
+    assert_equal("Sat", t2000.strftime("%a"))
+    assert_equal("Saturday", t2000.strftime("%A"))
+    assert_equal("Jan", t2000.strftime("%b"))
+    assert_equal("January", t2000.strftime("%B"))
+    assert_kind_of(String, t2000.strftime("%c"))
+    assert_equal("01", t2000.strftime("%d"))
+    assert_equal("00", t2000.strftime("%H"))
+    assert_equal("12", t2000.strftime("%I"))
+    assert_equal("001", t2000.strftime("%j"))
+    assert_equal("01", t2000.strftime("%m"))
+    assert_equal("00", t2000.strftime("%M"))
+    assert_equal("AM", t2000.strftime("%p"))
+    assert_equal("00", t2000.strftime("%S"))
+    assert_equal("00", t2000.strftime("%U"))
+    assert_equal("00", t2000.strftime("%W"))
+    assert_equal("6", t2000.strftime("%w"))
+    assert_equal("01/01/00", t2000.strftime("%x"))
+    assert_equal("00:00:00", t2000.strftime("%X"))
+    assert_equal("00", t2000.strftime("%y"))
+    assert_equal("2000", t2000.strftime("%Y"))
+    assert_equal("UTC", t2000.strftime("%Z"))
+    assert_equal("%", t2000.strftime("%%"))
+    assert_equal("0", t2000.strftime("%-S"))
 
-    assert_equal("", T2000.strftime(""))
-    assert_equal("foo\0bar\x0000\x0000\x0000", T2000.strftime("foo\0bar\0%H\0%M\0%S"))
-    assert_equal("foo" * 1000, T2000.strftime("foo" * 1000))
+    assert_equal("", t2000.strftime(""))
+    assert_equal("foo\0bar\x0000\x0000\x0000", t2000.strftime("foo\0bar\0%H\0%M\0%S"))
+    assert_equal("foo" * 1000, t2000.strftime("foo" * 1000))
 
     t = Time.mktime(2000, 1, 1)
     assert_equal("Sat", t.strftime("%a"))
+  end
 
+  def test_strftime_subsec
     t = Time.at(946684800, 123456.789)
     assert_equal("123", t.strftime("%3N"))
     assert_equal("123456", t.strftime("%6N"))
     assert_equal("123456789", t.strftime("%9N"))
     assert_equal("1234567890", t.strftime("%10N"))
     assert_equal("123456789", t.strftime("%0N"))
+  end
+
+  def test_strftime_sec
+    t = get_t2000.getlocal
     assert_equal("000", t.strftime("%3S"))
+  end
+
+  def test_strftime_seconds_from_epoch
+    t = Time.at(946684800, 123456.789)
     assert_equal("946684800", t.strftime("%s"))
     assert_equal("946684800", t.utc.strftime("%s"))
+  end
 
+  def test_strftime_zone
     t = Time.mktime(2001, 10, 1)
     assert_equal("2001-10-01", t.strftime("%F"))
+    assert_equal(Encoding::UTF_8, t.strftime("\u3042%Z").encoding)
+    assert_equal(true, t.strftime("\u3042%Z").valid_encoding?)
+  end
 
+  def test_strftime_flags
     t = Time.mktime(2001, 10, 1, 2, 0, 0)
     assert_equal("01", t.strftime("%d"))
     assert_equal("01", t.strftime("%0d"))
@@ -589,32 +728,61 @@ class TestTime < Test::Unit::TestCase
     assert_equal(" 2", t.strftime("%l"))
     assert_equal("02", t.strftime("%0l"))
     assert_equal(" 2", t.strftime("%_l"))
+  end
 
+  def test_strftime_invalid_flags
+    t = Time.mktime(2001, 10, 1, 2, 0, 0)
+    assert_equal("%4^p", t.strftime("%4^p"), 'prec after flag')
+  end
+
+  def test_strftime_year
+    t = Time.utc(1,1,4)
+    assert_equal("0001", t.strftime("%Y"))
+    assert_equal("0001", t.strftime("%G"))
+
+    t = Time.utc(0,1,4)
+    assert_equal("0000", t.strftime("%Y"))
+    assert_equal("0000", t.strftime("%G"))
+
+    t = Time.utc(-1,1,4)
+    assert_equal("-0001", t.strftime("%Y"))
+    assert_equal("-0001", t.strftime("%G"))
+  end
+
+  def test_strftime_weeknum
     # [ruby-dev:37155]
     t = Time.mktime(1970, 1, 18)
     assert_equal("0", t.strftime("%w"))
     assert_equal("7", t.strftime("%u"))
+  end
 
+  def test_strftime_ctrlchar
     # [ruby-dev:37160]
-    assert_equal("\t", T2000.strftime("%t"))
-    assert_equal("\t", T2000.strftime("%0t"))
-    assert_equal("\t", T2000.strftime("%1t"))
-    assert_equal("  \t", T2000.strftime("%3t"))
-    assert_equal("00\t", T2000.strftime("%03t"))
-    assert_equal("\n", T2000.strftime("%n"))
-    assert_equal("\n", T2000.strftime("%0n"))
-    assert_equal("\n", T2000.strftime("%1n"))
-    assert_equal("  \n", T2000.strftime("%3n"))
-    assert_equal("00\n", T2000.strftime("%03n"))
+    t2000 = get_t2000
+    assert_equal("\t", t2000.strftime("%t"))
+    assert_equal("\t", t2000.strftime("%0t"))
+    assert_equal("\t", t2000.strftime("%1t"))
+    assert_equal("  \t", t2000.strftime("%3t"))
+    assert_equal("00\t", t2000.strftime("%03t"))
+    assert_equal("\n", t2000.strftime("%n"))
+    assert_equal("\n", t2000.strftime("%0n"))
+    assert_equal("\n", t2000.strftime("%1n"))
+    assert_equal("  \n", t2000.strftime("%3n"))
+    assert_equal("00\n", t2000.strftime("%03n"))
+  end
 
+  def test_strftime_weekflags
     # [ruby-dev:37162]
-    assert_equal("SAT", T2000.strftime("%#a"))
-    assert_equal("SATURDAY", T2000.strftime("%#A"))
-    assert_equal("JAN", T2000.strftime("%#b"))
-    assert_equal("JANUARY", T2000.strftime("%#B"))
-    assert_equal("JAN", T2000.strftime("%#h"))
+    t2000 = get_t2000
+    assert_equal("SAT", t2000.strftime("%#a"))
+    assert_equal("SATURDAY", t2000.strftime("%#A"))
+    assert_equal("JAN", t2000.strftime("%#b"))
+    assert_equal("JANUARY", t2000.strftime("%#B"))
+    assert_equal("JAN", t2000.strftime("%#h"))
     assert_equal("FRIDAY", Time.local(2008,1,4).strftime("%#A"))
+  end
 
+  def test_strftime_rational
     t = Time.utc(2000,3,14, 6,53,"58.979323846".to_r) # Pi Day
     assert_equal("03/14/2000  6:53:58.97932384600000000000000000000",
                  t.strftime("%m/%d/%Y %l:%M:%S.%29N"))
@@ -634,11 +802,90 @@ class TestTime < Test::Unit::TestCase
                  t.strftime("%m/%d/%Y %l:%M:%S.%9N"))
     assert_equal("03/14/1592  6:53:58.97932384",
                  t.strftime("%m/%d/%Y %l:%M:%S.%8N"))
+  end
 
+  def test_strftime_far_future
     # [ruby-core:33985]
     assert_equal("3000000000", Time.at(3000000000).strftime('%s'))
   end
 
+  def test_strftime_too_wide
+    bug4457 = '[ruby-dev:43285]'
+    assert_raise(Errno::ERANGE, bug4457) {Time.now.strftime('%8192z')}
+  end
+
+  def test_strfimte_zoneoffset
+    t2000 = get_t2000
+    t = t2000.getlocal("+09:00:00")
+    assert_equal("+0900", t.strftime("%z"))
+    assert_equal("+09:00", t.strftime("%:z"))
+    assert_equal("+09:00:00", t.strftime("%::z"))
+    assert_equal("+09", t.strftime("%:::z"))
+
+    t = t2000.getlocal("+09:00:01")
+    assert_equal("+0900", t.strftime("%z"))
+    assert_equal("+09:00", t.strftime("%:z"))
+    assert_equal("+09:00:01", t.strftime("%::z"))
+    assert_equal("+09:00:01", t.strftime("%:::z"))
+  end
+
+  def test_strftime_padding
+    bug4458 = '[ruby-dev:43287]'
+    t2000 = get_t2000
+    t = t2000.getlocal("+09:00")
+    assert_equal("+0900", t.strftime("%z"))
+    assert_equal("+09:00", t.strftime("%:z"))
+    assert_equal("      +900", t.strftime("%_10z"), bug4458)
+    assert_equal("+000000900", t.strftime("%10z"), bug4458)
+    assert_equal("     +9:00", t.strftime("%_10:z"), bug4458)
+    assert_equal("+000009:00", t.strftime("%10:z"), bug4458)
+    assert_equal("  +9:00:00", t.strftime("%_10::z"), bug4458)
+    assert_equal("+009:00:00", t.strftime("%10::z"), bug4458)
+    assert_equal("+000000009", t.strftime("%10:::z"))
+    t = t2000.getlocal("-05:00")
+    assert_equal("-0500", t.strftime("%z"))
+    assert_equal("-05:00", t.strftime("%:z"))
+    assert_equal("      -500", t.strftime("%_10z"), bug4458)
+    assert_equal("-000000500", t.strftime("%10z"), bug4458)
+    assert_equal("     -5:00", t.strftime("%_10:z"), bug4458)
+    assert_equal("-000005:00", t.strftime("%10:z"), bug4458)
+    assert_equal("  -5:00:00", t.strftime("%_10::z"), bug4458)
+    assert_equal("-005:00:00", t.strftime("%10::z"), bug4458)
+    assert_equal("-000000005", t.strftime("%10:::z"))
+
+    bug6323 = '[ruby-core:44447]'
+    t = t2000.getlocal("+00:36")
+    assert_equal("      +036", t.strftime("%_10z"), bug6323)
+    assert_equal("+000000036", t.strftime("%10z"), bug6323)
+    assert_equal("     +0:36", t.strftime("%_10:z"), bug6323)
+    assert_equal("+000000:36", t.strftime("%10:z"), bug6323)
+    assert_equal("  +0:36:00", t.strftime("%_10::z"), bug6323)
+    assert_equal("+000:36:00", t.strftime("%10::z"), bug6323)
+    assert_equal("+000000:36", t.strftime("%10:::z"))
+    t = t2000.getlocal("-00:55")
+    assert_equal("      -055", t.strftime("%_10z"), bug6323)
+    assert_equal("-000000055", t.strftime("%10z"), bug6323)
+    assert_equal("     -0:55", t.strftime("%_10:z"), bug6323)
+    assert_equal("-000000:55", t.strftime("%10:z"), bug6323)
+    assert_equal("  -0:55:00", t.strftime("%_10::z"), bug6323)
+    assert_equal("-000:55:00", t.strftime("%10::z"), bug6323)
+    assert_equal("-000000:55", t.strftime("%10:::z"))
+  end
+
+  def test_strftime_invalid_modifier
+    t2000 = get_t2000
+    t = t2000.getlocal("+09:00")
+    assert_equal("%:y", t.strftime("%:y"), 'invalid conversion after : modifier')
+    assert_equal("%:0z", t.strftime("%:0z"), 'flag after : modifier')
+    assert_equal("%:10z", t.strftime("%:10z"), 'prec after : modifier')
+    assert_equal("%Ob", t.strftime("%Ob"), 'invalid conversion after locale modifier')
+    assert_equal("%Eb", t.strftime("%Eb"), 'invalid conversion after locale modifier')
+    assert_equal("%O0y", t.strftime("%O0y"), 'flag after locale modifier')
+    assert_equal("%E0y", t.strftime("%E0y"), 'flag after locale modifier')
+    assert_equal("%O10y", t.strftime("%O10y"), 'prec after locale modifier')
+    assert_equal("%E10y", t.strftime("%E10y"), 'prec after locale modifier')
+  end
+
   def test_delegate
     d1 = SimpleDelegator.new(t1 = Time.utc(2000))
     d2 = SimpleDelegator.new(t2 = Time.utc(2001))
@@ -685,4 +932,124 @@ class TestTime < Test::Unit::TestCase
       off += 0.1
     }
   end
+
+  def test_getlocal_dont_share_eigenclass
+    bug5012 = "[ruby-dev:44071]"
+
+    t0 = Time.now
+    class << t0; end
+    t1 = t0.getlocal
+
+    def t0.m
+      0
+    end
+
+    assert_raise(NoMethodError, bug5012) { t1.m }
+  end
+
+  def test_sec_str
+    bug6193 = '[ruby-core:43569]'
+    t = nil
+    assert_nothing_raised(bug6193) {t = Time.new(2012, 1, 2, 3, 4, "5")}
+    assert_equal(Time.new(2012, 1, 2, 3, 4, 5), t, bug6193)
+  end
+
+  def test_past
+    [
+      [-(1 << 100), 1, 1, 0, 0, 0],
+      [-4000, 1, 1, 0, 0, 0],
+      [-3000, 1, 1, 0, 0, 0],
+    ].each {|year, mon, day, hour, min, sec|
+      t = Time.utc(year, mon, day, hour, min, sec)
+      assert_equal(year, t.year)
+      assert_equal(mon, t.mon)
+      assert_equal(day, t.day)
+      assert_equal(hour, t.hour)
+      assert_equal(min, t.min)
+      assert_equal(sec, t.sec)
+    }
+  end
+
+  def test_1901
+    assert_equal(-0x80000001, Time.utc(1901, 12, 13, 20, 45, 51).tv_sec)
+    [
+      [1901, 12, 13, 20, 45, 50],
+      [1901, 12, 13, 20, 45, 51],
+      [1901, 12, 13, 20, 45, 52], # -0x80000000
+      [1901, 12, 13, 20, 45, 53],
+    ].each {|year, mon, day, hour, min, sec|
+      t = Time.utc(year, mon, day, hour, min, sec)
+      assert_equal(year, t.year)
+      assert_equal(mon, t.mon)
+      assert_equal(day, t.day)
+      assert_equal(hour, t.hour)
+      assert_equal(min, t.min)
+      assert_equal(sec, t.sec)
+    }
+  end
+
+  def test_1970
+    assert_equal(0, Time.utc(1970, 1, 1, 0, 0, 0).tv_sec)
+    [
+      [1969, 12, 31, 23, 59, 59],
+      [1970, 1, 1, 0, 0, 0],
+      [1970, 1, 1, 0, 0, 1],
+    ].each {|year, mon, day, hour, min, sec|
+      t = Time.utc(year, mon, day, hour, min, sec)
+      assert_equal(year, t.year)
+      assert_equal(mon, t.mon)
+      assert_equal(day, t.day)
+      assert_equal(hour, t.hour)
+      assert_equal(min, t.min)
+      assert_equal(sec, t.sec)
+    }
+  end
+
+  def test_2038
+    if no_leap_seconds?
+      assert_equal(0x80000000, Time.utc(2038, 1, 19, 3, 14, 8).tv_sec)
+    end
+    [
+      [2038, 1, 19, 3, 14, 7],
+      [2038, 1, 19, 3, 14, 8],
+      [2038, 1, 19, 3, 14, 9],
+      [2039, 1, 1, 0, 0, 0],
+    ].each {|year, mon, day, hour, min, sec|
+      t = Time.utc(year, mon, day, hour, min, sec)
+      assert_equal(year, t.year)
+      assert_equal(mon, t.mon)
+      assert_equal(day, t.day)
+      assert_equal(hour, t.hour)
+      assert_equal(min, t.min)
+      assert_equal(sec, t.sec)
+    }
+  end
+
+  def test_future
+    [
+      [3000, 1, 1, 0, 0, 0],
+      [4000, 1, 1, 0, 0, 0],
+      [1 << 100, 1, 1, 0, 0, 0],
+    ].each {|year, mon, day, hour, min, sec|
+      t = Time.utc(year, mon, day, hour, min, sec)
+      assert_equal(year, t.year)
+      assert_equal(mon, t.mon)
+      assert_equal(day, t.day)
+      assert_equal(hour, t.hour)
+      assert_equal(min, t.min)
+      assert_equal(sec, t.sec)
+    }
+  end
+
+  def test_getlocal_nil
+    now = Time.now
+    now2 = nil
+    now3 = nil
+    assert_nothing_raised {
+      now2 = now.getlocal
+      now3 = now.getlocal(nil)
+    }
+    assert_equal now2, now3
+    assert_equal now2.zone, now3.zone
+  end
 end
